import std.stdio;
import std.math;
import Vec3D;

class Matrix44
{
  double m[4][4] ;

  this(){
    m[0][0]=1.0; m[0][1]=0.0; m[0][2]=0.0; m[0][3]=0.0;
    m[1][0]=0.0; m[1][1]=1.0; m[1][2]=0.0; m[1][3]=0.0;
    m[2][0]=0.0; m[2][1]=0.0; m[2][2]=1.0; m[2][3]=0.0;
    m[3][0]=0.0; m[3][1]=0.0; m[3][2]=0.0; m[3][3]=1.0;
  }

  void setUnit(){
    m[0][0]=1.0; m[0][1]=0.0; m[0][2]=0.0; m[0][3]=0.0;
    m[1][0]=0.0; m[1][1]=1.0; m[1][2]=0.0; m[1][3]=0.0;
    m[2][0]=0.0; m[2][1]=0.0; m[2][2]=1.0; m[2][3]=0.0;
    m[3][0]=0.0; m[3][1]=0.0; m[3][2]=0.0; m[3][3]=1.0;
    return ;
  }

  void setRotate( double angle , vec3D axis ){
    double sinval, cosval;
    double x,y,z ;
    double d ;
    
    d=sqrt( axis.x*axis.x + axis.y*axis.y + axis.z*axis.z ) ;
    if( d==0.0 ){
      x=0.0 ;
      y=0.0 ;
      z=0.0 ;
    }else{
      x=axis.x/d ;
      y=axis.y/d ;
      z=axis.z/d ;
    }
    sinval = sin(angle) ;
    cosval = cos(angle) ;
    
    m[0][0] = x * x * (1.0 - cosval) +     cosval ;
    m[0][1] = x * y * (1.0 - cosval) - z * sinval ;
    m[0][2] = x * z * (1.0 - cosval) + y * sinval ;
    m[0][3] = 0.0 ;
    
    m[1][0] = x * y * (1.0 - cosval) + z * sinval ;
    m[1][1] = y * y * (1.0 - cosval) +     cosval ;
    m[1][2] = y * z * (1.0 - cosval) - x * sinval ;
    m[1][3] = 0.0 ;
    
    m[2][0] = x * z * (1.0 - cosval) - y * sinval ;
    m[2][1] = y * z * (1.0 - cosval) + x * sinval ;
    m[2][2] = z * z * (1.0 - cosval) +     cosval ;
    m[2][3] = 0.0 ;
    
    m[3][0] = 0.0 ;
    m[3][1] = 0.0 ;
    m[3][2] = 0.0 ;
    m[3][3] = 1.0 ;
    
    return ;
  }

  void setTranslate( vec3D dist ){
    m[0][0]=1.0 ; m[0][1]=0.0 ; m[0][2]=0.0 ; m[0][3]=dist.x ;
    m[1][0]=0.0 ; m[1][1]=1.0 ; m[1][2]=0.0 ; m[1][3]=dist.y ;
    m[2][0]=0.0 ; m[2][1]=0.0 ; m[2][2]=1.0 ; m[2][3]=dist.z ;
    m[3][0]=0.0 ; m[3][1]=0.0 ; m[3][2]=0.0 ; m[3][3]=1.0 ;
    return ;
  }
    
  void add( Matrix44 n ){
    for( int i=0 ; i<4 ; i++ ){
      for( int j=0 ; j<4 ; j++ ){
	m[i][j] += n.m[i][j] ;
      }
    }
    return ;
  }

  void sub( Matrix44 n ){
    for( int i=0 ; i<4 ; i++ ){
      for( int j=0 ; j<4 ; j++ ){
	m[i][j] -= n.m[i][j] ;
      }
    }
    return ;
  }

  void scale( double b ){
    for( int i=0 ; i<4 ; i++ ){
      for( int j=0 ; j<4 ; j++ ){
	m[i][j] *= b ;
      }
    }
    return ;
  }

  void iscale( double b ){
    for( int i=0 ; i<4 ; i++ ){
      for( int j=0 ; j<4 ; j++ ){
	m[i][j] /= b ;
      }
    }
    return ;
  }
    
  void mul( Matrix44 n ){
    double t[4][4] ;
  
    t[0][0] = m[0][0]*n.m[0][0] + m[0][1]*n.m[1][0] + m[0][2]*n.m[2][0] + m[0][3]*n.m[3][0] ;
    t[0][1] = m[0][0]*n.m[0][1] + m[0][1]*n.m[1][1] + m[0][2]*n.m[2][1] + m[0][3]*n.m[3][1] ;
    t[0][2] = m[0][0]*n.m[0][2] + m[0][1]*n.m[1][2] + m[0][2]*n.m[2][2] + m[0][3]*n.m[3][2] ;
    t[0][3] = m[0][0]*n.m[0][3] + m[0][1]*n.m[1][3] + m[0][2]*n.m[2][3] + m[0][3]*n.m[3][3] ;
    
    t[1][0] = m[1][0]*n.m[0][0] + m[1][1]*n.m[1][0] + m[1][2]*n.m[2][0] + m[1][3]*n.m[3][0] ;
    t[1][1] = m[1][0]*n.m[0][1] + m[1][1]*n.m[1][1] + m[1][2]*n.m[2][1] + m[1][3]*n.m[3][1] ;
    t[1][2] = m[1][0]*n.m[0][2] + m[1][1]*n.m[1][2] + m[1][2]*n.m[2][2] + m[1][3]*n.m[3][2] ;
    t[1][3] = m[1][0]*n.m[0][3] + m[1][1]*n.m[1][2] + m[1][2]*n.m[2][3] + m[1][3]*n.m[3][3] ;
    
    t[2][0] = m[2][0]*n.m[0][0] + m[2][1]*n.m[1][0] + m[2][2]*n.m[2][0] + m[2][3]*n.m[3][0] ;
    t[2][1] = m[2][0]*n.m[0][1] + m[2][1]*n.m[1][1] + m[2][2]*n.m[2][1] + m[2][3]*n.m[3][1] ;
    t[2][2] = m[2][0]*n.m[0][2] + m[2][1]*n.m[1][2] + m[2][2]*n.m[2][2] + m[2][3]*n.m[3][2] ;
    t[2][3] = m[2][0]*n.m[0][3] + m[2][1]*n.m[1][3] + m[2][2]*n.m[2][3] + m[2][3]*n.m[3][3] ;
    
    t[3][0] = m[3][0]*n.m[0][0] + m[3][1]*n.m[1][0] + m[3][2]*n.m[2][0] + m[3][3]*n.m[3][0] ;
    t[3][1] = m[3][0]*n.m[0][1] + m[3][1]*n.m[1][1] + m[3][2]*n.m[2][1] + m[3][3]*n.m[3][1] ;
    t[3][2] = m[3][0]*n.m[0][2] + m[3][1]*n.m[1][2] + m[3][2]*n.m[2][2] + m[3][3]*n.m[3][2] ;
    t[3][3] = m[3][0]*n.m[0][3] + m[3][1]*n.m[1][3] + m[3][2]*n.m[2][3] + m[3][3]*n.m[3][3] ;
    
    m[0][0]=t[0][0] ; m[0][1]=t[0][1] ; m[0][2]=t[0][2] ; m[0][3]=t[0][3] ;
    m[1][0]=t[1][0] ; m[1][1]=t[1][1] ; m[1][2]=t[1][2] ; m[1][3]=t[1][3] ;
    m[2][0]=t[2][0] ; m[2][1]=t[2][1] ; m[2][2]=t[2][2] ; m[2][3]=t[2][3] ;
    m[3][0]=t[3][0] ; m[3][1]=t[3][1] ; m[3][2]=t[3][2] ; m[3][3]=t[3][3] ;

    return ;
  }

  void print(){
    for( int i=0 ; i<4 ; i++ ){
      writef("%f,%f,%f,%f\n",m[i][0],m[i][1],m[i][2],m[i][3]) ;
    }
  }
}
