import std.stdio;
import std.math;
import Vec3D;
import Matrix44;

struct Quaternion {
  double n  = 1.0 ;
  double vx = 0.0 ;
  double vy = 0.0 ;
  double vz = 0.0 ;

  void set(double n, double vx, double vy, double vz ){
    this.n =n  ;
    this.vx=vx ;
    this.vy=vy ;
    this.vz=vz ;
  }

  void set(double n, vec3D *v ){
    this.n =n   ;
    this.vx=v.x ;
    this.vy=v.y ;
    this.vz=v.z ;
  }

  void set( Quaternion *q ){
    this.n =q.n  ;
    this.vx=q.vx ;
    this.vy=q.vy ;
    this.vz=q.vz ;
  }

  void setRotate( double angle, vec3D *axis ){
    if( angle<0.0 ){
      axis.inv() ;
      angle=-angle ;
    }
    
    double d ;
    double sinth=sin( angle/2.0 ) ;
    
    d=sqrt( axis.x*axis.x + axis.y*axis.y + axis.z*axis.z ) ;
    if( d==0.0 ){
      this.vx=0.0 ;
      this.vy=0.0 ;
      this.vz=1.0 ;
    }else{
      this.vx=axis.x/d ;
      this.vy=axis.y/d ;
      this.vz=axis.z/d ;
    }
    
    this.n   = cos( angle/2.0 ) ;
    this.vx *= sinth ;
    this.vy *= sinth ;
    this.vz *= sinth ;

    return ;
  }

  void setRotate( double angle, double vx, double vy, double vz ){
    if( angle<0.0 ){
      vx   =-vx ;
      vy   =-vy ;
      vz   =-vz ;
      angle=-angle ;
    }
    
    double d ;
    double sinth=sin( angle/2.0 ) ;
    
    d=sqrt( vx*vx + vy*vy + vz*vz ) ;
    if( d==0.0 ){
      this.vx=0.0 ;
      this.vy=0.0 ;
      this.vz=1.0 ;
    }else{
      this.vx=vx/d ;
      this.vy=vy/d ;
      this.vz=vz/d ;
    }
    
    this.n   = cos( angle/2.0 ) ;
    this.vx *= sinth ;
    this.vy *= sinth ;
    this.vz *= sinth ;

    return ;
  }

  void setAngle( double angle ){
    double d ;
    double sinth=sin( angle/2.0 ) ;
    
    d=sqrt( this.vx*this.vx + this.vy*this.vy + this.vz*this.vz ) ;
    if( d==0.0 ){
      this.vx=0.0 ;
      this.vy=0.0 ;
      this.vz=0.0 ;
    }else{
      this.vx=this.vx/d ;
      this.vy=this.vy/d ;
      this.vz=this.vz/d ;
    }
    
    this.n   = cos( angle/2.0 ) ;
    this.vx *= sinth ;
    this.vy *= sinth ;
    this.vz *= sinth ;

    return ;
  }

  void mul( Quaternion *qb ){
    double n,vx,vy,vz ;

    n = (this.n * qb.n ) - (this.vx * qb.vx) - (this.vy * qb.vy) - (this.vz * qb.vz) ;
    vx= (this.n * qb.vx) + (this.vx * qb.n)  + (this.vy * qb.vz) - (this.vz * qb.vy) ;
    vy= (this.n * qb.vy) + (this.vy * qb.n)  + (this.vz * qb.vx) - (this.vx * qb.vz) ;
    vz= (this.n * qb.vz) + (this.vz * qb.n)  + (this.vx * qb.vy) - (this.vy * qb.vx) ;
    
    this.n =n  ;
    this.vx=vx ;
    this.vy=vy ;
    this.vz=vz ;

    return ;
  }

  void conjugate(){
    this.vx = -this.vx ;
    this.vy = -this.vy ;
    this.vz = -this.vz ;
    return ;
  }

  double getAngle(){
    return( 2.0*acos(this.n) ) ;
  }

  vec3D getAxis(){
    vec3D retaxis ;
    if( getAngle()==0.0 ){
      retaxis.set(1.0,0.0,0.0) ;
    }else{
      retaxis.x=this.vx ;
      retaxis.y=this.vy ;
      retaxis.z=this.vz ;
      retaxis.normalize ;
    }
    return(retaxis) ;
  }

  void getMatrix( Matrix44 m ){
    m.m[0][0] = 1 - 2*( this.vy * this.vy + this.vz * this.vz ) ;
    m.m[0][1] =     2*( this.vx * this.vy - this.n  * this.vz ) ;
    m.m[0][2] =     2*( this.n  * this.vy + this.vx * this.vz ) ;
    m.m[0][3] = 0.0 ;

    m.m[1][0] =     2*( this.vx * this.vy + this.n  * this.vz ) ;
    m.m[1][1] = 1 - 2*( this.vx * this.vx + this.vz * this.vz ) ;
    m.m[1][2] =     2*( this.vy * this.vz - this.n  * this.vx ) ;
    m.m[1][3] = 0.0 ;
  
    m.m[2][0] =     2*( this.vx * this.vz - this.n  * this.vy ) ;
    m.m[2][1] =     2*( this.vy * this.vz + this.n  * this.vx ) ;
    m.m[2][2] = 1 - 2*( this.vx * this.vx + this.vy * this.vy ) ;
    m.m[2][3] = 0.0 ;

    m.m[3][0]=0.0 ; m.m[3][1]=0.0 ; m.m[3][2]=0.0 ; m.m[3][3]=1.0 ;
    return ;
  }

  void print(){
    writef("%f : %f, %f, %f\n",n,vx,vy,vz) ;
  }
}
    
