import win32.windows;
import std.stdio;
import std.string;
import std.utf;
import win32frame.writek;

static WindowBase[int] _WindowList ;
static WNDPROC[int]    _DefProc    ;

class WindowBase {
  HWND        hwnd;
  WNDCLASSEX  winc;
  string      name ;
  HMENU       hmenu ;
  int         finish_create ;

  this( string name ){
    this.winc.cbSize      = WNDCLASSEX.sizeof;
    this.winc.hInstance   = GetModuleHandle(null) ;
    this.winc.lpfnWndProc = cast(WNDPROC)(&_WndProc) ;
    this.hwnd             = null ;
    this.name             = name ;
    this.finish_create=0 ;
  }

  int setWNDCLASSEX( WNDCLASSEX winc){
    this.winc=winc ;
    return(0) ;
  }

  int registwinc(){
    if( !RegisterClassEx(&this.winc) ) return(-1) ;
    return(0) ;
  }
  
  int lowcreate(DWORD exstyle, DWORD style,
		int x, int y, int width, int height, HWND phwnd, HMENU hmenu){
    HWND rethwnd ;
    
    if( this.finish_create==0 ){
      this.hwnd=null ;
      rethwnd = CreateWindowEx(exstyle,
			       this.winc.lpszClassName,
			       cast(wchar*)(toUTF16(this.name ~ "\0")),
			       style,
			       x,y,width,height,
			       phwnd, hmenu, this.winc.hInstance , cast(LPVOID)(this)
			       );
      if( rethwnd == null ){
	writek("create failed %s %d,%d,%d,%d\n",name, x,y,width,height) ;
	return(-2) ;
      }else{
	if( this.hwnd == null ){
	  this.hwnd=rethwnd ;
	}
      }
      _WindowList[cast(int)(this.hwnd)]=this ;
      _DefProc[cast(int)(this.hwnd)] = cast(WNDPROC)(GetWindowLong(this.hwnd , GWL_WNDPROC)) ;
      SetWindowLong( this.hwnd, GWL_WNDPROC, cast(int)(&_WndSubProc)) ;
      this.finish_create=1 ;
    }
    this.hmenu=hmenu ;
    writek("create hwnd=%08x name=%s\n",cast(int)(this.hwnd),name) ;
    
    return(0) ;
  }

  HWND getHWND(){
    return( this.hwnd ) ;
  }
  
  HWND getParentHWND(){
    return( GetParent(this.hwnd) ) ;
  }
  
  HINSTANCE getInstance(){
    return( this.winc.hInstance ) ;
  }

  int setFont(HFONT hfont){
    if( this.hwnd!=null ){
      SendMessage( this.hwnd, WM_SETFONT, cast(WPARAM)(hfont), cast(LPARAM)(TRUE) );
      return(0) ;
    }
    return(-1) ;
  }

  LRESULT sendmsg(UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(this.hwnd,msg,wp,lp)) ;
  }

  LRESULT sendmsg(HWND tgthwnd, UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(tgthwnd,msg,wp,lp)) ;
  }

  LRESULT sendmsgtoparent(UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(this.getParentHWND(),msg,wp,lp)) ;
  }

  BOOL setWindowText(string str){
    int res = SendMessage( this.hwnd, WM_SETTEXT, 0, cast(LPARAM)(cast(wchar*)(toUTF16(str ~ "\0"))) ) ;
    if( res<=0 ){
      return(false) ;
    }
    return(true) ;
  }

  string getWindowTextUTF16(){
    char[] str = new char[GetWindowTextLength(this.hwnd)*2+2] ;
    GetWindowText(this.hwnd , cast(wchar*)(str) , str.length);
    str.length=str.length-2 ;
    return(str.idup) ;
  }
  
  string getWindowText(){
    wchar[] wstr   ; wstr.length = GetWindowTextLength(this.hwnd)+1 ;
    GetWindowText(this.hwnd , wstr.ptr , wstr.length);
    wstr.length=wstr.length-1 ;
    return( std.utf.toUTF8(wstr.idup) ) ;
  }
  
  LRESULT preproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return( this.proc(hwnd , msg , wp , lp) ) ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return( DefWindowProc(hwnd , msg , wp , lp) ) ;
  }

  LRESULT subproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return(-1) ;
  }

  LRESULT presubproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return(-1) ;
  }

  ~this(){
    writek("----WindowBase called destructor hwnd=%08x\n",this.hwnd) ;
    SendMessage( this.hwnd, WM_CLOSE  , 0, 0 ) ;
    _WindowList.remove(cast(int)(this.hwnd)) ;
    _DefProc.remove(cast(int)(this.hwnd)) ;
  }
}

extern (Windows) LRESULT _WndProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  
  if( msg==WM_DESTROY ){
    writek("WM_DESTROY hwnd=%08x , msg=%06x\n",cast(int)(hwnd),msg) ;
  }
  if( msg==WM_CLOSE ){
    writek("WM_CLOSE   hwnd=%08x , msg=%06x\n",cast(int)(hwnd),msg) ;
  }
  if( msg==WM_CREATE ){
    //WindowBase wb = cast(WindowBase)(*(cast(int*)(lp))) ;
    WindowBase wb = cast(WindowBase)(cast(void*)(*(cast(int*)(lp)))) ;
    wb.hwnd=hwnd ;
    _WindowList[cast(int)(hwnd)]=wb ;
  }
  if( (cast(int)(hwnd) in _WindowList) != null ){
    LRESULT res ;
    res = _WindowList[cast(int)(hwnd)].preproc(hwnd,msg,wp,lp) ;
    return( res ) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}

extern (Windows) LRESULT _WndSubProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  if( (cast(int)(hwnd) in _DefProc) != null ){
    _WindowList[cast(int)(hwnd)].presubproc(hwnd,msg,wp,lp) ;
    return(CallWindowProc( _DefProc[cast(int)(hwnd)] , hwnd, msg, wp, lp)) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}
